#include "platPorting.h"
#include "agentd.h"
#include "servConf.h"
#include "servRandom.h"

    /* server and agent information */

ServerData *serverData	 = NULL;
AgentSet  *agentSet	 = NULL;
CommandLine *commandLine = NULL;

/* sigterm_handler

   Purpose: This procedure handles the SIGTERM signal for all server processes.
*/

static sig_atomic_t sigtermInProgress = 0;

void sigterm_handler (int, SignalHandlers::ClientData)
{
    sigset_t mask;
    RandomFile::RandomFileErrors rc;

	/* bail if a SIGTERM is in progress -- this is required   */
	/* for IRIX where the "kill" statement below appears to   */
	/* be causing a SIGTERM interrupt even though SIGTERM is  */
	/* blocked automatically                                  */

    if (sigtermInProgress) {
	return;
    }

    sigtermInProgress = 1;

	/* block the SIGCHLD signal -- we don't save the old mask */
	/* since we are about to exit from the process and don't  */
	/* need to restore the old mask                           */

    sigemptyset (&mask);
    sigaddset (&mask, SIGCHLD);
    sigprocmask (SIG_BLOCK, &mask, (sigset_t *) NULL);

	/* propagate the signal to the rest of the process group */

    kill (-getpgrp(), SIGTERM);

	/* log a message that says we have received a SIGTERM */

    char pidString[16];
    sprintf (pidString, "%d", (int) getpid());

    char *argv[] = {
	"agentd: INFO: server process ",
	pidString,
	" received a SIGTERM: exiting",
	((char *) NULL)
    };

    serverData -> errorLog.error_app_cont (argv); 

	/* burn the random bits file if we are in agentd */

    if (getpid() == serverData -> agentdPid) {

	RandomFile randomFile (serverData -> randomFilename, serverData -> randomBits);

	if ((rc = randomFile.burnBits()) == RandomFile::e_OK) {

	    char *argv[] = {
		"agentd: INFO: burned random bits file \"",
		serverData -> randomFilename.value(),
		"\" before exit",
		((char *) NULL)
	    };

	    serverData -> errorLog.error_app_cont (argv); 

	} else {

	    char temp[16];
	    sprintf (temp, "%d %d", rc, errno);

	    char *argv[] = {
		"agentd: ERROR: unable to burn random bits file \"",
		serverData -> randomFilename.value(),
		"\" before exit",
		temp,
		((char *) NULL)
	    };

	    serverData -> errorLog.error_app_cont (argv); 
	}
    }

	/* exit -- all open file descriptors are automatically closed */

    exit (0);
}

/* sigio_handler

   Purpose: This procedure handles the SIGIO signal for all server processes.
*/

static sig_atomic_t sigioFired = 0;

int isSigioPending (void)
{
    sigset_t mask;
    sigpending (&mask);
    return (sigismember (&mask, SIGIO));
}
 
int hasSigioFired (void)
{
    int fired = sigioFired;
    sigioFired = 0;
    return (fired);
}

void pretendSigioFired (void) 
{
    sigioFired = 1;
}

void sigio_handler (int, SignalHandlers::ClientData)
{
    sigioFired = 1;
}

/* sighup_handler and sigprof_handler

   Purpose: This procedure handles the SIGHUP signal for all server processes.
*/

static sig_atomic_t sighupFired = 0; 

int isSighupPending (void)
{
    sigset_t mask;
    sigpending (&mask);
    return (sigismember (&mask, SIGPROF));
}

int hasSighupFired (void)
{
    int fired = sighupFired;
    sighupFired = 0;
    return (fired);
}

void pretendSighupFired (void) 
{
    sighupFired = 1;
}

void sighup_handler (int, SignalHandlers::ClientData)
{
	/* propagate the signal to the rest of the process group --  */
	/* we propagate a SIGPROF rather than a SIGHUP so that we do */
	/* not get into an infinite signal loop                      */

    kill (-getpgrp(), SIGPROF);
}

void sigprof_handler (int, SignalHandlers::ClientData)
{
    sighupFired = 1;
}

/* sigchld_handler

   Purpose: This procedure handles the SIGCHLD signal in the "agentd" process.
	    It detects when the "socketd" process dies and brings down the rest
            of the server processes.
*/

void sigchld_handler (int, SignalHandlers::ClientData)
{
    sigset_t mask;
    WAIT_STATUS_TYPE status;    

	/* block the SIGTERM signal -- we don't save the old mask */
	/* since we are about to exit from the process and don't  */
	/* need to restore the old mask                           */

    sigemptyset (&mask);
    sigaddset (&mask, SIGTERM);
    sigprocmask (SIG_BLOCK, &mask, (sigset_t *) NULL);

	/* remove the zombie */

    while (waitpid(0, (int *) &status, WNOHANG) > 0) {
	// empty
    }

	/* log an error message and then generate a SIGTERM to cleanup */

    char *argv[] = {
	"agentd: ERROR: socketd has failed so agentd is terminating",
	((char *) NULL)
    };

    serverData -> errorLog.error_app_cont (argv);
    sigterm_handler (SIGTERM, 0);
    abort_with_message ("should never get here");   // should never get here
}
